from __future__ import absolute_import, division, print_function

import os
import pickle
import shutil
from collections import Mapping, OrderedDict
from typing import Any, Text

import numpy as np
from six import string_types
from six.moves import cPickle, range, zip

from odin.fuel.databases import MmapDict, SQLiteDict
from odin.fuel.mmap_array import (MmapArray, MmapArrayWriter,
                                  read_mmaparray_header)
from odin.utils import (Progbar, UnitTimer, as_tuple, ctext, eprint,
                        flatten_list, get_file, is_callable, is_string, wprint)

__all__ = [
    'Dataset',
]

# ===========================================================================
# Helper
# ===========================================================================
def _infer_separator(path):
  all_sep = ('\t', ' ', ';', ',')
  with open(path, 'r') as f:
    line = f.readline()
    line = line.strip()
    orig_length = len(line)
    for s in all_sep:
      if s not in line:
        continue
      if 1 < len(line.split(s)) < orig_length:
        return s
    raise RuntimeError("CSV file with the first line: `%s`, "
                       "cannot match separator in known list: `%s`"
                       % (line, str(all_sep)))

_audio_ext = ('.3gp', '.aa', '.aac', '.aax', '.act', '.aiff',
              '.amr', '.ape', '.au', '.awb', '.dct', '.dss',
              '.dvf', '.flac', '.gsm', '.ivs', '.m4a', '.m4b',
              '.m4p', '.mmf', '.mp3', '.mpc', '.msv', '.nsf',
              '.ogg,', '.opus', '.raw', '.sln', '.tta', '.vox',
              '.wav', '.wma', '.wv', '.webm', '.sph', '.pcm')

_image_ext = ('.tif', '.tiff', '.gif', '.jpeg', '.jpg', '.jif',
              '.jfif', '.jp2', '.jpx', '.j2k', '.j2c', '.fpx',
              '.pcd', '.png', '.pdf')

_ignore_files = ('.DS_Store',)

def _parse_data_descriptor(path, read_only):
  """ Return mapping: name -> (dtype, shape, Data, path) """
  if not os.path.isfile(path):
    return None
  file_ext = os.path.splitext(path)[-1].lower()
  file_name = os.path.basename(path)
  # ====== ignore ====== #
  if os.path.basename(path) in _ignore_files:
    return None
  # ====== audio file ====== #
  if file_ext in _audio_ext:
    return [(file_name, ('audio', 'unknown', None, path))]
  # ====== image file ====== #
  if file_ext in _image_ext:
    return [(file_name, ('image', 'unknown', None, path))]
  # ====== text file .txt ====== #
  if file_ext in ('.txt',):
    return [(file_name, ('txt', 'unknown', None, path))]
  # ====== check if is csv file ====== #
  if file_ext in ('.csv', '.tsv'):
    sep = _infer_separator(path)
    data = []
    # read by manually open file much faster than numpy.genfromtxt
    with open(path, 'r') as f:
      for line in f:
        line = line[:-1]
        data.append(line.split(sep))
      data = np.array(data, dtype=str)
    return [('.'.join(file_name.split('.')[:-1]),
             ('csv', data.shape, data, path))]
  # ====== check if a file is Data ====== #
  try:
    dtype, shape = read_mmaparray_header(path)
    data = MmapArray(path)
    assert np.dtype(dtype) == data.dtype and shape == data.shape, \
      "Metadata mismatch for MmapArray"
    return [(file_name, (data.dtype, data.shape, data, path))]
  except Exception: # cannot read the header of MmapArray
    pass
  # ====== try to load pickle file if possible ====== #
  try: # try with unpickling
    with open(path, 'rb') as f:
      data = cPickle.load(f)
      shape_info = 0
      if hasattr(data, 'shape'):
        shape_info = data.shape
      elif hasattr(data, '__len__'):
        shape_info = len(data)
      return [(file_name, (str(data.dtype) if hasattr(data, 'dtype') else
                           type(data).__name__,
                           shape_info, data, path))]
  except cPickle.UnpicklingError:
    try: # try again with numpy load
      with open(path, 'rb') as f:
        data = np.load(f)
        return [(file_name,
        (str(data.dtype) if hasattr(data, 'dtype') else type(data).__name__,
         len(data) if hasattr(data, '__len__') else 0, data, path))]
    except Exception:
      pass
  # ====== load memmap dict ====== #
  try:
    data = MmapDict(path, read_only=read_only)
    return [(file_name, ('memdict', len(data), data, path))]
  except Exception as e:
    pass
  # ====== load SQLiteDict ====== #
  if '.db' in os.path.splitext(path)[1]:
    try:
      db = SQLiteDict(path, read_only=read_only)
      name = os.path.basename(path).replace('.db', '')
      return [(tab if tab != SQLiteDict._DEFAULT_TABLE else name,
               ('sqlite', len(db.set_table(tab)), db.as_table(tab), path))
              for tab in db.get_all_tables()]
    except Exception as e:
      pass
  # ====== unknown datatype ====== #
  return [(file_name, ('unknown', 'unknown', None, path))]


# ===========================================================================
# Datasets
# ===========================================================================
class Dataset(object):
  """ This Dataset can automatically parse memmap (created by MmapData),
  MmapDict, pickled dictionary and hdf5 files and keep tracking the changes.

  Any file name with "readme" prefix will be parsed as text and showed as
  readme.

  Support data type:
   - .txt or .csv files:
   -

  Note
  ----
  for developer: _data_map contains: name -> (dtype, shape, Data or pathtoData)
  readme included with the dataset should contain license information
  All the file with `.db` extension will be treat as SQLite data
  """

  __INSTANCES = {}

  def __new__(cls, *args, **kwargs):
    path = kwargs.get('path', None)
    if path is None:
      path = args[0]
    if not is_string(path):
      raise ValueError("`path` for Dataset must be string, but given "
                       "object with type: %s" % type(path))
    path = os.path.abspath(path)
    # Found old instance
    if path in Dataset.__INSTANCES:
      return Dataset.__INSTANCES[path]
    # new Dataset
    new_instance = super(Dataset, cls).__new__(cls)
    Dataset.__INSTANCES[path] = new_instance
    return new_instance

  def __init__(self, path, read_only=False, override=False):
    path = os.path.abspath(path)
    self.read_only = read_only
    self._readme_info = [ctext('README:', 'yellow'),
                         '------',
                         '  No information!']
    self._readme_path = None
    # flag to check cPickle called with protocol 2
    self._new_args_called = False
    # parse all data from path
    if path is not None:
      if override and os.path.exists(path) and os.path.isdir(path):
        shutil.rmtree(path)
        print('Overrided old dataset at path:', path)
      if os.path.isfile(path) and '.zip' in os.path.basename(path):
        self._load_archive(path,
            extract_path=path.replace(os.path.basename(path), ''),
            read_only=read_only)
      else:
        self._set_path(path, self.read_only)
    else:
      raise ValueError('Invalid path for Dataset: %s' % path)

  def _set_path(self, path, read_only):
    MAXIMUM_README_LINE = 25
    # all files are opened with default_mode=r+
    self._data_map = OrderedDict()
    self._path = os.path.abspath(path)
    self._default_hdf5 = os.path.basename(self._path) + '_default.h5'
    # svaed feeder info
    self._saved_indices = {}
    self._saved_recipes = {}
    # just make new dir
    if not os.path.exists(path):
      os.mkdir(path)
      os.mkdir(self.recipe_path)
      os.mkdir(self.index_path)
      return # not thing to do more
    elif not os.path.isdir(path):
      raise ValueError('Dataset path must be a folder.')
    # ====== Load all Data ====== #
    files = os.listdir(path)
    for fname in files:
      # found README
      if 'readme' == fname[:6].lower():
        readme_path = os.path.join(path, fname)
        with open(readme_path, 'r') as readme_file:
          readme = readme_file.readlines()[:MAXIMUM_README_LINE]
          readme = ['  ' + i[:-1] for i in readme if len(i) > 0 and i != '\n']
          readme.append(' => For more information: ' + readme_path)
          self._readme_info = [ctext('README:', 'yellow'),
                               '------'] + readme
          self._readme_path = readme_path
      # parse data
      data = _parse_data_descriptor(os.path.join(path, fname),
                                    read_only)
      if data is None: continue
      for key, d in data:
        if key in self._data_map:
          raise ValueError('Found duplicated data with follow info: '
                           '{}'.format(key))
        else:
          self._data_map[key] = d

  # ==================== Pickle ==================== #
  def __getstate__(self):
    if not self._new_args_called:
      raise RuntimeError(
          "You must use argument `protocol=cPickle.HIGHEST_PROTOCOL` "
          "when using `pickle` or `cPickle` to be able pickling Dataset.")
    self._new_args_called = False
    return self.path, self.read_only

  def __setstate__(self, states):
    path, read_only = states
    self._new_args_called = False
    self._set_path(path, read_only)

  def __getnewargs__(self):
    self._new_args_called = True
    return (self.path,)

  # ==================== archive loading ==================== #
  def _load_archive(self, path, extract_path, read_only):
    from zipfile import ZipFile, ZIP_DEFLATED
    try:
      zfile = ZipFile(path, mode='r', compression=ZIP_DEFLATED)
      allfile = zfile.namelist()
      # validate extract_path
      if not os.path.isdir(extract_path):
        raise ValueError('Extract path must be path folder, but path'
                         '={} is a file'.format(extract_path))
      extract_path = os.path.join(extract_path,
                                  os.path.basename(path).replace('.zip', ''))
      # found the extracted dir, use it
      if os.path.isdir(extract_path) and \
         set(os.listdir(extract_path)) == set(allfile):
        self._set_path(extract_path, read_only=read_only)
        return
      # decompress everything
      if not os.path.exists(extract_path):
        os.mkdir(extract_path)
      maxlen = max([len(i) for i in allfile])
      pb = Progbar(target=len(allfile), name="[Dataset] Loading Archive",
                   print_summary=True, print_report=True)
      for i, f in enumerate(allfile):
        zfile.extract(f, path=extract_path)
        pb['File'] = ('Unarchiving: %-' + str(maxlen) + 's') % f
        pb.add(1)
      # ====== finally set path ====== #
      self._set_path(extract_path, read_only=read_only)
    except IOError as e:
      raise IOError('Error loading archived dataset, path:{}, error:{}'
                    '.'.format(path, e))
    return None

  # ==================== properties ==================== #
  @property
  def basename(self):
    return os.path.basename(self.path)

  @property
  def path(self):
    return self._path

  @property
  def recipe_path(self):
    return os.path.join(self.path, 'recipe')

  @property
  def index_path(self):
    return os.path.join(self.path, 'index')

  @property
  def archive_path(self):
    """Return default archive path, which is:
        ../[dataset_name].zip
    """
    name = os.path.basename(self._path)
    return os.path.join(self._path, '..', name + '.zip')

  @property
  def md5_checksum(self):
    return self.get_md5_checksum()

  @property
  def size(self):
    """ return size in MegaByte"""
    size_bytes = 0
    for name, (dtype, shape, data, path) in self._data_map.items():
      try:
        size_bytes += os.path.getsize(path) # in bytes
      except Exception as e:
        eprint("Cannot acquire file size information, file: %s; error: %s"
               % (str(name), str(e)))
    return size_bytes / 1024. / 1024.

  def __len__(self):
    """ Return total number of data """
    return len(self._data_map)

  def __iter__(self):
    return self.items()

  def items(self):
    for name in self._data_map.keys():
      yield name, self.__getitem__(name)

  def iterinfo(self):
    """Return iteration of: (dtype, shape, loaded_data, path)"""
    for name, (dtype, shape, data, path) in self._data_map.items():
      yield (dtype, shape, self.__getitem__(name), path)

  def keys(self):
    """
    Return
    ------
    name of all Data
    """
    return self._data_map.keys()

  def values(self):
    """
    Return
    ------
    (dtype, shape, data, path) of Data
    """
    for k in self._data_map.keys():
      yield self.__getitem__(k)

  def archive(self):
    from zipfile import ZipFile, ZIP_DEFLATED
    path = self.archive_path
    zfile = ZipFile(path, mode='w', compression=ZIP_DEFLATED)

    files = set([_[-1] for _ in self._data_map.values()])

    prog = Progbar(target=len(files), name="[Dataset] Archiving",
                   print_report=True, print_summary=True)
    maxlen = max([len(os.path.basename(i)) for i in files])
    for i, f in enumerate(files):
      zfile.write(f, os.path.basename(f))
      prog['Data'] = ('Archiving: %-' + str(maxlen) + 's') \
          % os.path.basename(f)
      prog.add(1)
    zfile.close()
    return path

  # ==================== Data management ==================== #
  def copy(self, destination,
           indices_filter=None, data_filter=None,
           override=False):
    """ Copy the dataset to a new folder and closed
    the old dataset

    """
    from distutils.dir_util import copy_tree
    read_only = self.read_only
    raise NotImplementedError

  def flush(self):
    for dtype, shape, data, path in self._data_map.values():
      if hasattr(data, 'flush'):
        data.flush()
      elif data is not None: # Flush pickling data
        with open(path, 'wb') as f:
          cPickle.dump(data, f, protocol=cPickle.HIGHEST_PROTOCOL)

  def close(self, name=None):
    # ====== close all Data ====== #
    if name is None: # close all files
      for name, (dtype, shape, data, path) in list(self._data_map.items()):
        if hasattr(data, 'close'):
          data.close()
        del data
        del self._data_map[name]
      # close all external indices and recipes
      for name, ids in self._saved_indices.items():
        ids.close()
      self._saved_indices.clear()
      for name, rcp in list(self._saved_recipes.items()):
        del rcp
      self._saved_recipes.clear()
      # Check if exist global instance
      if self.path in Dataset.__INSTANCES:
        del Dataset.__INSTANCES[self.path]
    # ====== close a particular file ====== #
    elif name in self._data_map:
      (dtype, shape, data, path) = self._data_map[name]
      if dtype == 'sqlite':
        data.sqlite.close()
      elif hasattr(data, 'close'):
        data.close()
      del data
      del self._data_map[name]

  # ==================== Some info ==================== #
  def __contains__(self, key):
    return key in self._data_map

  def find_prefix(self, feat_name, prefix):
    """ Specialized method for searching for Data or NoSQL
    with prefix, for example `prefix='indices'`:
      - `indices_%s_%s` % (feat1_name, feat2, ...)
    if no indices found, return the default indices with
    name 'indices'
    """
    indices = self[prefix] if prefix in self else None
    for key in self.keys():
      if prefix == key[:len(prefix)] and '_' + feat_name in key:
        indices = self[key]
    if indices is None:
      raise RuntimeError("Cannot find prefix: '%s' for feature with name: '%s', "
                         "all available name with given prefix are: %s" %
                         (prefix, feat_name, ','.join([k for k in self.keys()
                                                       if prefix == k[:len(k)]])
                         ))
    return indices

  def __getitem__(self, key):
    if is_string(key):
      if key not in self._data_map:
        raise KeyError('%s not found in this dataset' % key)
      dtype, shape, data, path = self._data_map[key]
      return path if data is None else data
    raise ValueError('Only accept key type is string.')

  def get(self, key, default=None):
    if key in self._data_map:
      return self.__getitem__(key)
    return default

  def __setitem__(self, name: Text, value: Any):
    """
    Parameters
    ----------
    key : str or tuple
        if tuple is specified, it contain the key and the datatype
        which must be "memmap", "hdf5"
        for example: ds[('X', 'hdf5')] = numpy.ones((8, 12))
    """
    assert isinstance(name, string_types), \
      "name must be given as string types."
    path = os.path.join(self.path, name)
    with open(path, 'wb') as f:
      pickle.dump(value, f)
    self._data_map[name] = (
      value.dtype if hasattr(value, 'dtype') else str(type(value)),
      value.shape if hasattr(value, 'shape') else 'unknown',
      value, path)


  def get_md5_checksum(self, excluded_name=[]):
    from odin.utils.crypto import md5_checksum
    md5_text = ''
    all_data_items = {i: j
                      for i, j in self._data_map.items()
                      if i not in excluded_name}
    for name, (dtype, shape, data, path) in sorted(all_data_items.items(),
                                                   key=lambda x: x[0]):
      md5_text += md5_checksum(path)
    return md5_text

  def __str__(self):
    padding = '  '
    # NOTE: each element in the list is one line
    s = ['==========  ' +
         ctext('Dataset:%s Total:%d Size:%.2f(MB)', 'magenta') %
         (self.path, len(self._data_map), self.size) +
         '  ==========']
    s += self._readme_info
    s += [ctext('DATA:', 'yellow'),
          '----']
    # ====== Find longest string ====== #
    longest_name = 0
    longest_shape = 0
    longest_dtype = 0
    longest_file = 0
    print_info = []
    for name, (dtype, shape, data, path) in sorted(self._data_map.items()):
      shape = data.shape if hasattr(data, 'shape') else shape
      longest_name = max(len(name), longest_name)
      longest_dtype = max(len(str(dtype)), longest_dtype)
      longest_shape = max(len(str(shape)), longest_shape)
      longest_file = max(len(str(path)), longest_file)
      print_info.append([name, dtype, shape, path])
    # ====== return print string ====== #
    format_str = (padding + '%-' + str(longest_name + 2) + 's  '
                  '%-' + str(longest_dtype) + 's' + ctext(':', 'yellow') +
                  '%-' + str(longest_shape) + 's  '
                  'path:%-' + str(longest_file) + 's')
    for name, dtype, shape, path in print_info:
      s.append(format_str % ('"%s"' % name, dtype, shape, path))
    # ====== add recipes info ====== #
    for name, recipe in self._saved_recipes.items():
      s.append(ctext('(Recipe) ', 'yellow') + '"%s"' % name)
      for rcp in recipe:
        rcp = str(rcp)
        s.append('\n'.join([padding + line
                            for line in rcp.split('\n')]))
    # ====== add indices info ====== #
    for name, index in self._saved_indices.items():
      s.append(ctext('(Index) ', 'yellow') + '"%s"' % name)
      s.append(padding + str(index))
      name, (start, end) = next(index.items())
      s.append(padding + 'Sample: "%s %d-%d"' % (name, start, end))
    return '\n'.join(s)

  @property
  def readme(self):
    """ return text string of README of this dataset """
    if self._readme_path is not None:
      with open(self._readme_path, 'r') as f:
        readme = f.read()
    else:
      readme = self._readme_info[-1]
    return readme



from __future__ import absolute_import, division, print_function

import inspect
from typing import List, Optional, Text

import numpy as np
import tensorflow as tf
from six import string_types
from tensorflow_probability.python import distributions as tfd
from tensorflow_probability.python.distributions import kullback_leibler
from tensorflow_probability.python.internal import (assert_util, prefer_static,
                                                    tensorshape_util)
from tensorflow_probability.python.layers.internal import \
    distribution_tensor_coercible

from odin.bay import distributions as obd
from odin.bay.distributions.negative_binomial_disp import NegativeBinomialDisp
from odin.bay.distributions.zero_inflated import ZeroInflated

__all__ = ['concat_distribution']

# ===========================================================================
# Helpers
# ===========================================================================
# must hand define all the parameters here
# NOTE: this list is to be updated, or a smarter solution for automatically
# mining all the parameters
dist_params = {
    # complex
    obd.Independent: ['distribution', 'reinterpreted_batch_ndims'],
    ZeroInflated: ['count_distribution', 'inflated_distribution'],
    obd.MixtureSameFamily: ['mixture_distribution', 'components_distribution'],
    # Exponential
    obd.Gamma: ['concentration', 'rate'],
    # Gaussians
    obd.Normal: ['loc', 'scale'],
    obd.LogNormal: ['loc', 'scale'],
    obd.MultivariateNormalDiag: ['loc', 'scale'],
    obd.MultivariateNormalTriL: ['loc', 'scale'],
    obd.MultivariateNormalFullCovariance: ['loc', 'scale'],
    # Count
    NegativeBinomialDisp: ['loc', 'disp'],
    obd.NegativeBinomial: ['total_count', 'logits_parameter'],
    obd.Poisson: ['log_rate_parameter'],
    # Binary and probability
    obd.Gumbel: ['loc', 'scale'],
    obd.Bernoulli: ['logits_parameter'],
    obd.Dirichlet: ['concentration'],
    obd.Beta: ['concentration1', 'concentration0'],
    obd.OneHotCategorical: ['logits_parameter'],
    obd.Categorical: ['logits_parameter'],
    # others
    obd.Laplace: ['loc', 'scale'],
    obd.Wishart: ['df', 'scale'],
    obd.Uniform: ['low', 'high'],
    obd.Multinomial: ['total_count', 'logits_parameter'],
    obd.Deterministic: ['loc', 'atol', 'rtol'],
    obd.VectorDeterministic: ['loc', 'atol', 'rtol'],
}

for dist_type, attr_names in dist_params.items():
  assert isinstance(attr_names, (tuple, list)) and all(
      isinstance(name, string_types) for name in attr_names), \
        "Error defining parameters of distributions"
  assert isinstance(dist_type, type) and issubclass(dist_type, obd.Distribution),\
        "Error defining parameters of distributions"
  assert all(hasattr(dist_type, name) for name in attr_names), \
        "Error defining parameters of distributions"


# ===========================================================================
# Main code
# ===========================================================================
def _find_axis_for_stack(dists, given_axis):
  # check event shape is consistent
  if given_axis is not None:
    return int(given_axis)

  event_shape = dists[0].event_shape
  batch_shape = dists[0].batch_shape

  assertions = []
  for d in dists[1:]:
    assertions.append(tf.assert_equal(event_shape, d.event_shape))
    assertions.append(tf.assert_equal(batch_shape.ndims, d.batch_shape.ndims))

  with tf.control_dependencies(assertions):
    axis = []
    for d in dists:
      shape = d.batch_shape
      for ax, (i, j) in enumerate(zip(batch_shape, shape)):
        if i != j:
          axis.append(ax)
    if len(axis) == 0:
      return 0
    assert len(set(axis)) == 1, \
      "Multiple dimensions are found to be different among the distributions, "\
        "expect only 1 different dimension."
    return axis[0]


def concat_distribution(dists: List[tfd.Distribution],
                        axis: Optional[int] = None,
                        validate_args: bool = False,
                        allow_nan_stats: bool = True,
                        name: Optional[Text] = None) -> tfd.Distribution:
  """ This layer create a new `Distribution` by concatenate parameters of
  multiple distributions of the same type along given `axis`

  Note
  ----
  If your distribution is the output from
  `tensorflow_probability.DistributionLambda`, this function will remove all
  the keras tracking ultilities, for better solution checkout
  `odin.networks.distribution_util_layer.ConcatDistribution`
  """
  if not isinstance(dists, (tuple, list)):
    dists = [dists]
  if len(dists) == 1:
    return dists[0]
  if len(dists) == 0:
    raise ValueError("No distributions were given")
  axis = _find_axis_for_stack(dists, given_axis=axis)

  t = type(dists[0])
  is_keras_output = False
  # _TensorCoercible will messing up with the parameters of the
  # distribution
  if issubclass(t, distribution_tensor_coercible._TensorCoercible):
    is_keras_output = True
    t = type.mro(t)[2]
    assert issubclass(t, tfd.Distribution) and not issubclass(
        t, distribution_tensor_coercible._TensorCoercible)

  # no more distribution, tensor of parameters is return during the
  # recursive operator
  if issubclass(t, tf.Tensor):
    if dists[0].shape.ndims == 0:
      return dists[0]  # TODO: better solution here
    return tf.concat(dists, axis=axis)
  elif issubclass(t, obd.Distribution):
    pass  # continue with all distribution parameters
  else:
    return dists[0]

  # get all params for concatenate
  if t not in dist_params:
    raise RuntimeError("Unknown distribution of type '%s' for concatenation" %
                       str(t))
  params_name = dist_params[t]

  # start concat the params
  params = {}
  for p in params_name:
    attrs = [getattr(d, p) for d in dists]
    is_method = False
    if inspect.ismethod(attrs[0]):
      attrs = [a() for a in attrs]
      is_method = True
    if is_method and '_parameter' == p[-10:]:
      p = p[:-10]

    params[p] = concat_distribution(attrs, axis=axis)

  # extra arguments
  if name is not None:
    params['name'] = name
  args = inspect.getfullargspec(t.__init__).args
  if 'allow_nan_stats' in args:
    params['allow_nan_stats'] = allow_nan_stats
  if 'validate_args' in args:
    params['validate_args'] = validate_args
  dist = t(**params)

  return dist



# -*- coding: utf-8 -*-
""""
author: 'Omid Sadjadi, Timothee Kheyrkhah'
email: 'omid.sadjadi@nist.gov'
"""
import time
import warnings
from numbers import Number

import numpy as np
from scipy.linalg import cholesky, eigh, inv, solve, svd
from six import string_types

from odin.backend import calc_white_mat, length_norm
from odin.ml.base import BaseEstimator, Evaluable, TransformerMixin
from odin.ml.scoring import (VectorNormalizer, compute_class_avg,
                             compute_within_cov)
from odin.utils import unique


def logdet(A):
  u = cholesky(A)
  y = 2 * np.log(np.diag(u)).sum()
  return y

class PLDA(BaseEstimator, TransformerMixin, Evaluable):
  """ Probabilistic LDA

  Parameters
  ----------
  n_phi : int
    number of dimension for the latent space

  centering : bool (default: True)
    mean normalization the data before EM

  wccn : bool (default: True)
    within class covariance normalization before EM

  unit_length : bool (default: True)
    normalize vector length of each sample to 1 before EM

  n_iter : {integer, 'auto'}
    if 'auto', keep iterating until no more improvement (i.e. reduction in `sigma` value)
    compared to the `improve_threshold`

  improve_threshold : scalar
    Only used in case `n_iter='auto'`

  labels : {list of string, or None} (default: None)
    labels information for `evaluate` method

  seed : int
    random seed for reproducibility

  verbose : int (default: 0)
    verbose level, 0 for turning off all logging activities,
    1 for basics notification, 2 for fitting progress.
    if `2`, compute log-likelihood during fitting EM, this will
    significantly slows down the process, only suggested for debugging

  Attributes
  ----------
  Sigma_ : [feat_dim, feat_dim]
  Phi_ : [feat_dim, n_phi]
  Sb_ : [feat_dim, feat_dim]
  St_ : [feat_dim, feat_dim]
  Lambda : []
  Uk : []
  Q_hat : []
  X_model_ : [num_class, feat_dim]
    class-dependence feature vectors
  """

  def __init__(self, n_phi=None,
               centering=True, wccn=True, unit_length=True,
               n_iter='auto', improve_threshold=1e-1,
               labels=None, dtype='float64', random_state=None,
               verbose=0):
    super(PLDA, self).__init__()
    # ====== check n_phi ====== #
    if n_phi is not None:
      n_phi = int(n_phi)
    self.n_phi_ = n_phi
    # ====== check num_iter ====== #
    if isinstance(n_iter, string_types):
      n_iter = n_iter.lower()
      assert n_iter == 'auto', 'Invalid `n_iter` value: %s' % n_iter
    elif isinstance(n_iter, Number):
      assert n_iter > 0, "`n_iter` must greater than 0, but given: %d" % n_iter
    self.n_iter_ = n_iter
    self.improve_threshold_ = float(improve_threshold)
    # ====== other ====== #
    self.feat_dim_ = None
    self._labels = labels
    self.verbose_ = int(verbose)
    # for normalization
    self._normalizer = VectorNormalizer(
        centering=centering, wccn=wccn, unit_length=unit_length,
        lda=False, concat=False)
    self._dtype = np.dtype(dtype)
    # ====== check random state ====== #
    if random_state is None:
      self._rand_state = np.random.RandomState(None)
    elif isinstance(random_state, Number):
      self._rand_state = np.random.RandomState(seed=random_state)
    elif isinstance(random_state, np.random.RandomState):
      self._rand_state = random_state
    else:
      raise ValueError("Invalid argument for `random_state`: %s" % str(random_state))
    # Attributes
    self.Sigma_ = None
    self.Phi_ = None
    self.Sb_ = None
    self.St_ = None

  # ==================== properties ==================== #
  @property
  def dtype(self):
    return self._dtype

  @property
  def feat_dim(self):
    return self.feat_dim_

  @property
  def normalizer(self):
    return self._normalizer

  @property
  def labels(self):
    return self._labels

  @property
  def num_classes(self):
    return len(self._labels)

  @property
  def is_fitted(self):
    if not hasattr(self, 'Lambda_') or \
    not hasattr(self, 'Uk_') or \
    not hasattr(self, 'Q_hat_') or \
    not hasattr(self, 'X_model_'):
      return False
    return True

  # ==================== Pickling ==================== #
  def __getstate__(self):
    if not self.is_fitted:
      raise RuntimeError("The PLDA have not been fitted, nothing to pickle!")
    return (self.n_phi_, self.n_iter_, self.feat_dim_, self._labels, self.verbose_,
            self._normalizer, self._dtype, self._rand_state,
            self.Sigma_, self.Phi_, self.Sb_, self.St_,
            self.Lambda_, self.Uk_, self.Q_hat_, self.X_model_)

  def __setstate__(self, states):
    (self.n_phi_, self.n_iter_, self.feat_dim_, self._labels, self.verbose_,
     self._normalizer, self._dtype, self._rand_state,
     self.Sigma_, self.Phi_, self.Sb_, self.St_,
     self.Lambda_, self.Uk_, self.Q_hat_, self.X_model_) = states

  # ==================== helpers ==================== #
  def initialize(self, X, labels):
    feat_dim = X.shape[1]
    if self.feat_dim is None or self._num_classes is None:
      self.feat_dim_ = int(feat_dim)
      if self._labels is None:
        self._labels = labels
      if self.feat_dim <= self.n_phi_:
        raise RuntimeError("`feat_dim=%d` must be greater than `n_phi=%d`" %
          (self.feat_dim, self.n_phi_))
      # ====== initialize ====== #
      # covariance matrix of the residual term
      # self.Sigma_ = 1. / self.feat_dim * np.eye(self.feat_dim, dtype=self.dtype)
      self.Sigma_ = (1. / self.feat_dim * np.eye(self.feat_dim) +
                    self._rand_state.randn(self.feat_dim, self.feat_dim)
                    ).astype(self.dtype)
      # self.Sigma_ = np.cov(X.T).astype(self.dtype)
      # self.Sigma_ = (np.cov(X.T) +
      #               self._rand_state.randn(self.feat_dim, self.feat_dim)
      #               ).astype(self.dtype)
      # self.Sigma_ = 100 * self._rand_state.randn(
      #     self.feat_dim, self.feat_dim).astype(self.dtype)
      # factor loading matrix (Eignevoice matrix) [feat_dim, n_phi]
      # self.Phi_ = np.r_[np.eye(self.n_phi_),
      #                  np.zeros((self.feat_dim - self.n_phi_, self.n_phi_))]
      # self.Phi_ = self._rand_state.randn(self.feat_dim, self.n_phi_).astype(self.dtype)
      self.Phi_ = self.normalizer.transform(
          self._rand_state.randn(self.n_phi_, self.feat_dim)
      ).T.astype(self.dtype)
      self.Sb_ = np.zeros((self.feat_dim, self.feat_dim), dtype=self.dtype)
      self.St_ = np.zeros((self.feat_dim, self.feat_dim), dtype=self.dtype)
    # ====== validate the dimension ====== #
    if self.feat_dim != feat_dim:
      raise ValueError("Mismatch the input feature dimension, %d != %d" %
        (self.feat_dim, feat_dim))
    if self.num_classes != len(labels):
      raise ValueError("Mismatch the number of output classes, %d != %d" %
        (self.num_classes, len(labels)))

  # ==================== sklearn ==================== #
  def _update_caches(self):
    # ====== update cached matrices for scoring ====== #
    iSt = inv(self.St_) # [feat_dim, feat_dim]
    iS = inv(self.St_ - np.dot(np.dot(self.Sb_, iSt), self.Sb_))
    Q = iSt - iS # [feat_dim, feat_dim]
    P = np.dot(np.dot(iSt, self.Sb_), iS) # [feat_dim, feat_dim]
    U, s, V = svd(P, full_matrices=False)
    self.Lambda_ = np.diag(s[:self.n_phi_]) # [n_phi, n_phi]
    self.Uk_ = U[:, :self.n_phi_] # [feat_dim, n_phi]
    self.Q_hat_ = np.dot(np.dot(self.Uk_.T, Q), self.Uk_) # [n_phi, n_phi]

  def fit_maximum_likelihood(self, X, y):
    # ====== preprocessing ====== #
    if isinstance(X, (tuple, list)):
      X = np.asarray(X)
    elif "odin.fuel" in str(type(X)):
      X = X[:]
    if isinstance(y, (tuple, list)):
      y = np.asarray(y)
    # ====== normalizing and initializing ====== #
    X = self.normalizer.fit(X, y).transform(X)
    classes = np.unique(y)
    self.initialize(X, labels=classes)
    # ====== ml ====== #
    Sw = compute_within_cov(X, y, classes)
    self.St_ = np.cov(X.T)
    self.Sb_ = self.St_ - Sw
    # ====== the default class_avg ====== #
    self._update_caches()
    model_vecs = compute_class_avg(X, y, classes=classes)
    self.X_model_ = np.dot(model_vecs, self.Uk_)
    return self

  def fit(self, X, y):
    """
    Parameters
    ----------
    X : [num_samples, feat_dim]
    y : [num_samples]
    """
    # ====== preprocessing ====== #
    if isinstance(X, (tuple, list)):
      X = np.asarray(X)
    elif "odin.fuel" in str(type(X)):
      X = X[:]
    if isinstance(y, (tuple, list)):
      y = np.asarray(y)
    assert X.shape[0] == y.shape[0], \
        "Number of samples mismatch in `X` and `y`, %d != %d" % \
        (X.shape[0], y.shape[0])
    # ====== normalize and initialize ====== #
    y_counts = np.bincount(y) # sessions per speaker
    classes = np.unique(y)
    X = self.normalizer.fit(X, y).transform(X)
    self.initialize(X, labels=classes)
    # ====== Initializing ====== #
    F = np.zeros((self.num_classes, self.feat_dim))
    for clz in np.unique(y):
      # Speaker indices
      F[clz, :] = X[y == clz, :].sum(axis=0)
    if self.verbose_ > 0:
      print('Re-estimating the Eigenvoice subspace with {} factors ...'.format(self.n_phi_))
    X_sqr = np.dot(X.T, X)
    # ====== iteration ====== #
    iter = 0
    last_llk_value = None
    while True:
      e_time = time.time()
      # expectation
      Ey, Eyy = self.expectation_plda(F, y_counts)
      e_time = time.time() - e_time
      # maximization
      m_time = time.time()
      self.maximization_plda(X, X_sqr, F, Ey, Eyy)
      m_time = time.time() - m_time
      # log-likelihood
      llk = 'None'
      llk_value = None
      if self.verbose_ > 1 or isinstance(self.n_iter_, string_types):
        llk_value = self.compute_llk(X)
        llk = '%.2f' % llk_value
      if self.verbose_ > 0:
        print('#iter:%-3d \t [llk = %s] \t [E-step = %.2f s] [M-step = %.2f s]' %
              (iter + 1, llk, e_time, m_time))
      # check breaking condition
      iter += 1
      if isinstance(self.n_iter_, Number):
        if iter >= self.n_iter_:
          break
      elif iter > 2 and last_llk_value is not None:
        if llk_value - last_llk_value < self.improve_threshold_:
          break
      last_llk_value = llk_value
    # ====== Update the eigenvoice space ====== #
    self.Sb_ = self.Phi_.dot(self.Phi_.T)
    self.St_ = self.Sb_ + self.Sigma_
    # ====== the default class_avg ====== #
    self._update_caches()
    model_vecs = compute_class_avg(X, y, classes=classes)
    self.X_model_ = np.dot(model_vecs, self.Uk_)

  def expectation_plda(self, F, cls_counts):
    """
    Parameters
    ----------
    F : [num_classes, feat_dim]
    cls_count : [num_classes]
    """
    # computes the posterior mean and covariance of the factors
    num_classes = F.shape[0]
    Eyy = np.zeros(shape=(self.n_phi_, self.n_phi_))
    Ey_clz = np.zeros(shape=(num_classes, self.n_phi_))
    # initialize common terms to save computations
    uniqFreqs = unique(cls_counts, keep_order=True)
    n_uniq = len(uniqFreqs)
    invTerms = np.empty(shape=(n_uniq, self.n_phi_, self.n_phi_))
    PhiT_invS = solve(self.Sigma_.T, self.Phi_).T # [n_phi, feat_dim]
    PhiT_invS_Phi = np.dot(PhiT_invS, self.Phi_) # [n_phi, n_phi]
    I = np.eye(self.n_phi_)

    for ix in range(n_uniq):
      nPhiT_invS_Phi = uniqFreqs[ix] * PhiT_invS_Phi
      invTerms[ix] = inv(I + nPhiT_invS_Phi)

    for clz in range(num_classes):
      num_samples = cls_counts[clz]
      PhiT_invS_y = np.dot(PhiT_invS, F[clz, :])
      idx = np.flatnonzero(uniqFreqs == num_samples)[0]
      Cyy = invTerms[idx]
      Ey_clz[clz, :] = np.dot(Cyy, PhiT_invS_y)
      Eyy += num_samples * Cyy

    Eyy += np.dot((Ey_clz * cls_counts[:, None]).T, Ey_clz)
    return Ey_clz, Eyy

  def compute_llk(self, X):
    """
    Parameters
    ----------
    X : [num_samples, feat_dim]
    """
    num_samples = X.shape[0]
    S = np.dot(self.Phi_, self.Phi_.T) + self.Sigma_ # [feat_dim, feat_dim]
    llk = -0.5 * (self.feat_dim * num_samples * np.log(2 * np.pi) +
                  num_samples * logdet(S) +
                  np.sum(X * solve(S, X.T).T))
    return llk

  def maximization_plda(self, X, X_sqr, F, Ey, Eyy):
    """
    ML re-estimation of the Eignevoice subspace and the covariance of the
    residual noise (full).

    Paremters
    ---------
    X : [num_samples, feat_dim]
    X_cov : [feat_dim, feat_dim]
    F : [num_classes, feat_dim]
    Ey : [num_classes, n_phi]
    Eyy : [n_phi, n_phi]
    """
    num_samples = X.shape[0]
    Ey_FT = np.dot(Ey.T, F) # [n_phi, feat_dim]
    self.Phi_ = solve(Eyy.T, Ey_FT).T # [feat_dim, n_phi]
    self.Sigma_ = 1. / num_samples * (X_sqr - np.dot(self.Phi_, Ey_FT))

  def transform(self, X):
    if not self.is_fitted:
      raise RuntimeError("This model hasn't been fitted!")
    # ====== check X ====== #
    if isinstance(X, (tuple, list)):
      X = np.asarray(X)
    elif "odin.fuel" in str(type(X)):
      X = X[:]
    # ====== transform into latent space ====== #
    X_norm = self.normalizer.transform(X)
    X_project = np.dot(X_norm, self.Uk_) # [num_samples, n_phi]
    return X_project
    # return np.dot(X_project, self.Q_hat_)
    # h = np.dot(X_project, self.Q_hat_) * X_project
    # return h

  def predict_log_proba(self, X, X_model=None):
    """
    Parameters
    ----------
    X : [num_samples, feat_dim]
    X_model : [num_classes, feat_dim]
      if None, use class average extracted based on fitted data

    Return
    ------
    log-probabilities matrix [num_samples, num_classes]
    """
    if not self.is_fitted:
      raise RuntimeError("This model hasn't been fitted!")
    # ====== check X_model ====== #
    if X_model is None:
      X_model = self.X_model_
    else:
      # [num_classes, n_phi]
      X_model = np.dot(self.normalizer.transform(X_model), self.Uk_)
    if X_model.shape[0] != self.num_classes:
      warnings.warn("The model matrix contains %d classes, but the "
                    "fitted number of classes is %d" %
                    (X_model.shape[0], self.num_classes))
    # ====== check X ====== #
    if isinstance(X, (tuple, list)):
      X = np.asarray(X)
    elif "odin.fuel" in str(type(X)):
      X = X[:]
    # ====== transform the input matrices ====== #
    X = np.dot(self.normalizer.transform(X), self.Uk_) # [num_samples, n_phi]
    # [num_classes, 1]
    score_h1 = np.sum(np.dot(X_model, self.Q_hat_) * X_model, axis=1, keepdims=True)
    # [num_samples, 1]
    score_h2 = np.sum(np.dot(X, self.Q_hat_) * X, axis=1, keepdims=True)
    # [num_samples, num_classes]
    score_h1h2 = 2 * np.dot(X, np.dot(X_model, self.Lambda_).T)
    # [num_samples, num_classes]
    scores = score_h1h2 + score_h1.T + score_h2
    return scores



from __future__ import absolute_import, division, print_function

import copy
import inspect
import numbers
from collections import defaultdict
from functools import wraps

import numpy as np
import scipy as sp
import tensorflow as tf
from six.moves import builtins
from tensorflow.python.ops import init_ops

from odin.utils import as_tuple, is_number, is_same_shape, is_string, uuid


# ===========================================================================
# Helper
# ===========================================================================
def _normalize_axis(axis, ndim):
  if axis is None:
    return None
  if isinstance(axis, (tuple, list)):
    return tuple([a % ndim if a is not None else a
            for a in axis])
  return axis % ndim

# ===========================================================================
# Normalization
# ===========================================================================
def length_norm(x, axis=-1, epsilon=1e-12, ord=2):
  """ L2-normalization (or vector unit length normalization)

  Parameters
  ----------
  x : array
  axis : int
  ord : int
    order of norm (1 for L1-norm, 2 for Frobenius or Euclidean)
  """
  ord = int(ord)
  if ord not in (1, 2):
    raise ValueError("only support `ord`: 1 for L1-norm; 2 for Frobenius or Euclidean")
  if ord == 2:
    x_norm = tf.sqrt(tf.maximum(tf.reduce_sum(x ** 2, axis=axis, keepdims=True),
                                epsilon))
  else:
    x_norm = tf.maximum(tf.reduce_sum(tf.abs(x), axis=axis, keepdims=True),
                        epsilon)
  return x / x_norm

def calc_white_mat(X):
  """ calculates the whitening transformation for cov matrix X
  """
  return tf.linalg.cholesky(tf.linalg.inv(X))

def log_norm(x, axis=1, scale_factor=10000, eps=1e-8):
  """ Seurat log-normalize
  y = log(X / (sum(X, axis) + epsilon) * scale_factor)

  where `log` is natural logarithm
  """
  eps = tf.cast(eps, x.dtype)
  return tf.math.log1p(
      x / (tf.reduce_sum(x, axis=axis, keepdims=True) + eps) * scale_factor)

def delog_norm(x, x_sum=1, scale_factor=10000):
  """ This perform de-log normalization of `log_norm` values
  if `x_sum` is not given (i.e. default value 1), then all the
  """
  return (tf.exp(x) - 1) / scale_factor * (x_sum + EPS)

# ===========================================================================
# Conversion
# ===========================================================================
def logreduceexp(x, reduction_function=tf.reduce_mean, axis=None,
                 name=None):
  """ log-reduction-exp over axis to avoid overflow and underflow

  Parameters
  ----------
  `x` : [nb_sample, feat_dim]
  `axis` should be features dimension
  """
  with tf.name_scope(name, "logreduceexp"):
    x_max = tf.reduce_max(x, axis=axis, keepdims=True)
    y = tf.log(
        reduction_function(tf.exp(x - x_max), axis = axis, keepdims=True)) + x_max
    return tf.squeeze(y)

def logsumexp(x, axis=-1, name=None):
  """
  `x` : [nb_sample, feat_dim]
  `axis` should be features dimension
  """
  # ====== tensorflow ====== #
  with tf.name_scope(name, 'logsumexp', [x]):
    xmax = tf.reduce_max(x, axis=axis, keepdims=True)
    y = xmax + tf.log(tf.reduce_sum(tf.exp(x - xmax), axis=axis, keepdims=True))
  return y

def to_llh(x, name=None):
  ''' Convert a matrix of probabilities into log-likelihood
  :math:`LLH = log(prob(data|target))`
  '''
  with tf.name_scope(name, "log_likelihood", [x]):
    x /= tf.reduce_sum(x, axis=-1, keepdims=True)
    x = tf.clip_by_value(x, EPS, 1 - EPS)
    return tf.log(x)

def to_llr(x, name=None):
  ''' Convert a matrix of probabilities into log-likelihood ratio
  :math:`LLR = log(\\frac{prob(data|target)}{prob(data|non-target)})`
  '''
  with tf.name_scope(name, "log_likelihood_ratio", [x]):
    nb_classes = x.shape.as_list()[-1]
    new_arr = []
    for j in range(nb_classes):
      scores_copy = tf.transpose(tf.gather(tf.transpose(x),
                                 [i for i in range(nb_classes) if i != j]))
      scores_copy -= tf.expand_dims(x[:, j], axis=-1)
      new_arr.append(-logsumexp(scores_copy, 1))
    return tf.concat(new_arr, axis=-1) + np.log(13)

def to_nonzeros(x, value):
  x = tf.where(tf.equal(x, 0.), tf.zeros_like(x) + value, x)
  return x

def to_sample_weights(indices, weights, name=None):
  """ Convert indices or one-hot matrix and
  give weights to sample weights for training """
  with tf.name_scope(name, "to_sample_weights", [indices]):
    # ====== preprocess indices ====== #
    ndim = len(indices.shape)
    if ndim <= 1: # indices vector
      indices = tf.cast(indices, dtype=tf.int64)
    else:
      indices = tf.argmax(indices, axis=-1)
    # ====== prior weights ====== #
    if isinstance(weights, (tuple, list, np.ndarray)):
      prior_weights = tf.constant(weights, dtype=floatX,
                            name="prior_weights")
    # ====== sample weights ====== #
    weights = tf.gather(prior_weights, indices)
  return weights

# ===========================================================================
# Allocation
# ===========================================================================
def tril(m, k=0, name=None):
  """
  Lower triangle of an array.

  Return a copy of an array with elements above the `k`-th diagonal zeroed.

  Parameters
  ----------
  m : array_like, shape (M, N)
      Input array.
  k : int, optional
      Diagonal above which to zero elements.  `k = 0` (the default) is the
      main diagonal, `k < 0` is below it and `k > 0` is above.

  Returns
  -------
  tril : ndarray, shape (M, N)
      Lower triangle of `m`, of same shape and data-type as `m`.
  """
  with tf.name_scope(name, 'LowerTriangle'):
    if k == 0:
      return tf.matrix_band_part(input=m, num_lower=-1, num_upper=0, name=name)
    if k < 0:
      return tf.subtract(m,
        tf.matrix_band_part(input=m, num_lower=np.abs(k) - 1, num_upper=-1),
        name=name)
    # k > 0
    return tf.matrix_band_part(input=m, num_lower=-1, num_upper=k, name=name)

def tril_indices(n, k=0, name=None):
  """ Similar as `numpy.tril_indices`
  @Author: avdrher
  https://github.com/GPflow/GPflow/issues/439

  Return the indices for the lower-triangle of an (n, m) array.

  Parameters
  ----------
  n : int
      The row dimension of the arrays for which the returned
      indices will be valid.
  k : int, optional
      Diagonal above which to zero elements.  `k = 0` (the default) is the
      main diagonal, `k < 0` is below it and `k > 0` is above.

  Returns
  -------
  inds : tuple of arrays
      The indices for the triangle. The returned tuple contains two arrays,
      each with the indices along one dimension of the array.

  """
  with tf.name_scope(name, "LowerTriangleIndices"):
    M1 = tf.tile(tf.expand_dims(tf.range(n), axis=0), [n, 1])
    M2 = tf.tile(tf.expand_dims(tf.range(n), axis=1), [1, n])
    mask = tf.transpose((M1 - M2) >= -k)
    ix1 = tf.boolean_mask(M2, mask)
    ix2 = tf.boolean_mask(M1, mask)
    return ix1, ix2

def prior2weights(prior, exponential=False,
                  min_value=0.1, max_value=None,
                  norm=False):
  """ TODO: finish this

  Parameters
  ----------
  prior: numpy.ndarray [nb_classes,]
      probabilty values of each classes prior,
      sum of all prior must be equal to 1.
  exponential: bool
  min_value: bool
  max_value: bool
  norm: bool
      if True, normalize output weights to sum up to 1.
  """
  # idea is the one with highest prior equal to 1.
  # and all other classes is the ratio to this prior
  prior = np.array(prior).ravel()
  prior = 1 / prior * np.max(prior)
  # print(prior)
  if exponential:
    prior = sorted([(i, p) for i, p in enumerate(prior)],
                   key=lambda x: x[-1], reverse=False)
    alpha = interp.expIn(n=len(prior), power=10)
    prior = {i: a * p for a, (i, p) in zip(alpha, prior)}
    prior = np.array([prior[i] for i in range(len(prior))]) + 1
  # ====== rescale everything within max_value ====== #
  if min_value is not None and max_value is not None:
    min_value = float(min_value)
    max_value = float(max_value)
    prior = (max_value - min_value) * (prior - np.min(prior)) \
        / (np.max(prior) - np.min(prior)) + min_value
  # ====== normaize by ====== #
  if norm:
    prior = prior / np.sum(prior)
  return prior

def entropy(p, name=None):
  """Return simple calculation of discrete Shanon entropy"""
  with tf.name_scope(name, "entropy"):
    return -tf.reduce_sum(p * tf.log(p))

def upsample(x, scale, axes, method='nn', name=None):
  """
  Parameters
  ----------
  scale: int, list of int
      scaling up factor
  axes: int, list of int
      the axes of tensor which the upsampling method will be applied
  method: str, int
      'nn' for nearest neighbor (e.g. [1, 2] => [1, 1, 2, 2]),
      'pad' for padding within the tensor. 'pad_margin' do padding
      in the margin of the tensor. 'repeat' simple algorithm for
      repeating the element (e.g. [1, 2] => [1, 2, 1, 2])
  """
  with tf.name_scope(name, "Upsample"):
    method = method.lower()
    input_shape = tf.shape(x)
    input_shape_int = x.shape.as_list()
    ndims = x.shape.ndims
    # normalize all negative axes
    if axes is None:
      raise ValueError("axes cannot be None.")
    axes = [1, 2] if axes is None else \
        [i % ndims for i in as_tuple(axes)]
    sorted(axes)
    # make scale a tuple
    scale = as_tuple(scale, N=len(axes), t=int)
    # mapping from axis -> scale
    scale_map = defaultdict(lambda: 1)
    scale_map.update([(i, j) for i, j in zip(axes, scale)])
    # create final output_shape
    output_shape = [input_shape[i] * scale_map[i] for i in range(ndims)]
    # ====== Nearest neighbor method ====== #
    if method == 'nn':
      # tensorflow only support for tile <= 6-D tensor
      if ndims >= 6:
        raise ValueError('upsample with NN mode does not support rank >= 6 tensor.')
      elif ndims + len(axes) > 6:
        for a in axes:
          x = upsample(x, scale_map[a], axes=a, method='nn')
      else:
        # repeat the tensor
        x = dimshuffle(x, pattern=list(range(ndims)) + ['x'] * len(axes))
        x = repeat(x, scale, axes=[i for i in range(ndims, ndims + len(axes))])
        # transpose it back to the right shape
        axes_map = {i: j for i, j in zip(axes, range(ndims, ndims + len(axes)))}
        new_axes = []
        for i in range(ndims):
          if i not in axes_map:
            new_axes.append(i)
          else:
            new_axes += [i, axes_map[i]]
        x = tf.transpose(x, perm=new_axes)
        x = reshape(x, output_shape)
    # ====== pading_margin ====== #
    elif method.lower() == 'pad_margin':
      paddings = [[0, 0] if i not in axes else
          [tf.cast(tf.ceil(input_shape[i] * (scale_map[i] - 1) / 2), 'int32'),
           tf.cast(tf.floor(input_shape[i] * (scale_map[i] - 1) / 2), 'int32')]
          for i in range(ndims)]
      x = tf.pad(x, paddings=paddings, mode='CONSTANT')
    # ====== pading ====== #
    elif method == 'pad':
      raise NotImplementedError
      # x = tf.scatter_nd(indices, x, shape=output_shape)
    # ====== repeat ====== #
    elif method == 'repeat':
      x = repeat(x, n=scale, axes=axes)
    # ====== no support ====== #
    else:
      raise ValueError("No support for method='%s'" % method)
    # ====== add_shape ====== #
    return set_shape(x, shape=[
        s * scale_map[i] if is_number(s) else None
        for i, s in enumerate(input_shape_int)])

# ===========================================================================
# Linear Algebra
# ===========================================================================
def dot(x, y, name=None):
  """ Theano-style dot product
  For 2-D arrays it is equivalent to matrix multiplication,
  and for 1-D arrays to inner product of vectors (without complex conjugation).
  For N dimensions it is a sum product over the last axis of a and
  the second-to-last of b

  NOTE: this behavior is the same in `numpy.dot` as well but
  hasn't replicated in `tensorflow.matmul`

  Example
  -------
   (2, 3).(4, 3, 5) => (2, 4, 5)
   (2, 3, 4).(4, 5) => (2, 3, 5)
   (2, 3, 4).(5, 4, 6) => (2, 3, 5, 6)
  """
  with tf.name_scope(name, "DotProduct"):
    shapeX = [tf.shape(x)[i] if d is None else d
              for i, d in enumerate(x.shape.as_list())]
    shapeY = [tf.shape(y)[i] if d is None else d
              for i, d in enumerate(y.shape.as_list())]
    ndimX = x.shape.ndims
    ndimY = y.shape.ndims
    if ndimX > 2:
      x = tf.reshape(x, (-1, shapeX[-1]))
    if ndimY > 2:
      y_dims = list(range(ndimY))
      y_dims = [y_dims.pop(-2)] + y_dims
      y = tf.transpose(y, perm=y_dims)
      y = tf.reshape(y, (shapeY[-2], -1))
      outshapeY = [shapeY[i] for i in y_dims[1:]]
    else:
      outshapeY = [shapeY[-1]]
    # calculate dot product and desire shape
    output_shape = shapeX[:-1] + outshapeY
    output = tf.reshape(tf.matmul(x, y), output_shape)
  return output

def batched_dot(x, y, name=None):
  """Batchwise dot product.
  This function computes the dot product between the two tensors,
  by iterating over the first dimension.
  """
  with tf.name_scope(name, "BatchedDot"):
    shapeX = [tf.shape(x)[i] if d is None else d
              for i, d in enumerate(x.shape.as_list())]
    shapeY = [tf.shape(y)[i] if d is None else d
              for i, d in enumerate(y.shape.as_list())]
    ndimX = x.shape.ndims
    ndimY = y.shape.ndims
    # same as dot but one more batch dimension
    if ndimX > 2 + 1:
      x = tf.reshape(x, (-1, np.prod(shapeX[1:-1]), shapeX[-1]))
    if ndimY > 2 + 1:
      y_dims = list(range(ndimY))
      y_dims = [y_dims.pop(0), y_dims.pop(-2)] + y_dims
      y = tf.transpose(y, perm=y_dims)
      outshapeY = [shapeY[i] for i in y_dims[2:]]
      y = tf.reshape(y, (-1, shapeY[-2], np.prod(outshapeY)))
    else:
      outshapeY = [shapeY[-1]]
    # calculate dot product and desire shape
    output_shape = shapeX[:-1] + outshapeY
    output = tf.reshape(tf.matmul(x, y),
                        [i if i is not None else -1 for i in output_shape],
                        name=name)
    return output

def switch(condition, then_expression, else_expression, name=None):
  with tf.name_scope(name, 'switch'):
    if condition.dtype != tf.bool:
      condition = tf.cast(condition, 'bool')
    x_shape = copy.copy(then_expression.shape)
    # tensorflow require the last dimension of 3 variables is equal, too
    # it is irrelevant since condition can have shape[-1] = 1
    cond_ndims = condition.shape.ndims
    if cond_ndims > 1 and condition.shape[-1] != x_shape[-1]:
      cond_shape = tf.shape(condition)
      condition = tf.reshape(condition,
          [cond_shape[i] for i in range(cond_ndims - 1)])
    x = tf.where(condition, then_expression, else_expression)
    x.set_shape(x_shape)
    return x

def apply_mask(x, mask, name="ApplyMask"):
  """
  x : 3D tensor
  mask : 2D tensor

  Example
  -------
  >>> Input: [128, 500, 120]
  >>> Mask:  [1, 1, 0]
  >>> Output: [128, 500, 0]
  """
  return tf.mul(x, tf.expand_dims(mask, -1), name=name)


# ===========================================================================
# Shape manipulation
# ===========================================================================
def reshape(x, shape, name='Reshape'):
  """ More flexible version of reshape operation

  Example
  -------
  x.shape = [25, 08, 12]
  reshape(shape=([1], [2], [0]))
  => x.shape = (08, 12, 25)
  """
  input_shape = x.shape.as_list()
  new_shape = []
  for i in shape:
    if i is None:
      new_shape.append(-1)
    elif isinstance(i, (list, tuple)):
      new_shape.append(input_shape[i[0]])
    else:
      new_shape.append(i)
  new_shape = tuple([-1 if i is None else i
                     for i in new_shape])
  return tf.reshape(x, new_shape, name=name)

def dimshuffle(x, pattern, name='Dimshuffle'):
  """Transpose dimensions.

  pattern should be a tuple or list of
  dimension indices, e.g. [0, 2, 1].
  """
  with tf.name_scope(name):
    x = tf.transpose(x, perm=[i for i in pattern if i != 'x'])
    # insert new dimension
    for i, p in enumerate(pattern):
      if p == 'x':
        x = tf.expand_dims(x, i)
  return x

def flatten(x, outdim=1, name='Flatten'):
  """ Keep all the original dimension until `outdim - 1`
  """
  with tf.name_scope(name):
    if outdim == 1:
      return tf.reshape(x, [-1], name=name)
    input_shape = [tf.shape(x)[i] if d is None else d
                   for i, d in enumerate(x.shape.as_list())]
    other_shape = tuple([input_shape[i] for i in range(outdim - 1)])
    n = 1
    for i in input_shape[(outdim - 1):]:
      n = n * i
    output_shape = other_shape + (n,)
    return tf.reshape(x, output_shape)

def repeat(x, n, axes=None, name="Repeat"):
  """ Repeat a N-D tensor.

  If x has shape (s1, s2, s3) and axes=(1, -1), the output
  will have shape (s1, s2 * n[0], s3 * n[1]).

  Parameters
  ----------
  n : {int, list of int}
    each number of repeatation according to the axes
  axes : {int, list or int}
    all axes for repeating
  """
  if axes is not None:
    ndim = x.shape.ndims
    if not isinstance(axes, (tuple, list)):
      axes = (axes,)
    axes = _normalize_axis(axes, ndim)
    n = as_tuple(n, len(axes))
    return tf.tile(x, multiples=[n[axes.index(i)] if i in axes else 1
                                 for i in range(ndim)],
                   name=name)
  else:
    n = int(n)
    return tf.tile(x, multiples=[n for i in range(ndim)],
                   name=name)


# ===========================================================================
# Statistics
# ===========================================================================
def var(x, axes=None, keepdims=False, name="Variance"):
  with tf.name_scope(name):
    axes = _normalize_axis(axes, x.shape.ndims)
    x = tf.cast(x, floatX)
    m = tf.reduce_mean(x, axis=axes, keepdims=True)
    devs_squared = tf.square(x - m)
    return tf.reduce_mean(devs_squared, axis=axes, keepdims=keepdims)


def std(x, axes=None, keepdims=False, name="Std"):
  with tf.name_scope(name):
    return tf.sqrt(var(x, axes=axes, keepdims=keepdims), name=name)


def renorm_rms(X, axis=1, target_rms=1.0, name="RescaleRMS"):
  """ Scales the data such that RMS of the features dimension is 1.0
  scale = sqrt(x^t x / (D * target_rms^2)).

  NOTE
  ----
  by defaults, assume the features dimension is `1`
  """
  with tf.name_scope(name):
    D = tf.sqrt(tf.cast(tf.shape(X)[axis], X.dtype.base_dtype))
    l2norm = tf.sqrt(tf.reduce_sum(X ** 2, axis=axis, keepdims=True))
    X_rms = l2norm / D
    X_rms = tf.where(tf.equal(X_rms, 0.),
                     x=tf.ones_like(X_rms, dtype=X_rms.dtype.base_dtype),
                     y=X_rms)
    return target_rms * X / X_rms



from __future__ import print_function
import os
import sys
import base64
import shutil
import inspect
from six import add_metaclass
from zipfile import ZipFile, ZIP_DEFLATED
from abc import ABCMeta, abstractproperty

import numpy as np

from odin.utils.crypto import unzip_aes
from odin.utils import (get_file, get_script_path, ctext, get_datasetpath,
                        string_normalize)
from odin.fuel.dataset import Dataset

# ===========================================================================
# Helper
# ===========================================================================
def parse_dataset(dataset_name):
  fn_norm = lambda x: string_normalize(x, lower=True,
                                       remove_whitespace='',
                                       remove_non_alphanumeric=True)
  dataset_name = fn_norm(dataset_name)
  # ====== get all datasets ====== #
  all_datasets = []
  for name, member in inspect.getmembers(sys.modules[__name__]):
    if (isinstance(member, type) and issubclass(member, DataLoader)) \
    or 'load_' == name[:5]:
      name = name.replace('load_', '')
      name = fn_norm(name)
      all_datasets.append((name, member))
  # ====== search for best match ====== #
  match_dataset = [name for name, _ in all_datasets
                   if name == dataset_name]
  if len(match_dataset) == 0:
    match_dataset = [name for name, _ in all_datasets
                     if dataset_name == name[:len(dataset_name)]]
  if len(match_dataset) == 0:
    raise RuntimeError("Cannot find dataset with name '%s', "
      "all available datasets are: %s" % (
          dataset_name,
          ', '.join([i[0] for i in all_datasets])))
  if len(match_dataset) > 1:
    raise RuntimeError("Found multiple dataset for name: '%s', "
      "all the candidates are: %s" % (
          dataset_name,
          ', '.join([i for i in match_dataset])))
  # ====== extract the found dataset ====== #
  match_dataset = match_dataset[0]
  dataset = [ds for name, ds in all_datasets
             if name == match_dataset][0]
  if 'load_' in match_dataset:
    return match_dataset()
  return dataset.load()

def unzip_folder(zip_path, out_path, remove_zip=True):
  if '.zip' not in zip_path:
    raise ValueError(".zip extension must be in the zip_path.")
  if not os.path.exists(zip_path):
    raise ValueError("Cannot find zip file at path: %s" % zip_path)
  try:
    zf = ZipFile(zip_path, mode='r', compression=ZIP_DEFLATED)
    zf.extractall(path=out_path)
    zf.close()
  except Exception:
    shutil.rmtree(out_path)
    import traceback; traceback.print_exc()
  finally:
    if remove_zip:
      os.remove(zip_path)

@add_metaclass(ABCMeta)
class DataLoader(object):
  ORIGIN = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzLw==\n'
  BASE_DIR = get_datasetpath(root='~')

  def __init__(self):
    super(DataLoader, self).__init__()

  @classmethod
  def md5(clazz, ext=''):
    return None

  @classmethod
  def get_name(clazz, ext=''):
    name = clazz.__name__
    name = name if ext is None or len(ext) == 0 \
        else '_'.join([name, ext])
    return name

  @classmethod
  def get_zip_path(clazz, ext=''):
    return os.path.join(DataLoader.BASE_DIR,
                        clazz.get_name(ext) + '.zip')

  @classmethod
  def get_ds_path(clazz, ext=''):
    return os.path.join(DataLoader.BASE_DIR, clazz.get_name(ext))

  @classmethod
  def get_link(clazz, ext=''):
    name = clazz.get_name(ext) + '.zip'
    path = base64.decodebytes(DataLoader.ORIGIN).decode() + name
    return path

  @classmethod
  def load(clazz, ext='', override=False):
    return clazz.get_dataset(ext=ext, override=override)

  @classmethod
  def get_dataset(clazz, ext='', override=False):
    # ====== all path ====== #
    name = clazz.get_name(ext) + '.zip'
    path = base64.decodebytes(DataLoader.ORIGIN).decode() + name
    zip_path = clazz.get_zip_path(ext)
    out_path = clazz.get_ds_path(ext)
    # ====== check out_path ====== #
    if os.path.isfile(out_path):
      raise RuntimeError("Found a file at path: %s, we need a folder "
                         "to unzip downloaded files." % out_path)
    elif os.path.isdir(out_path):
      if override or len(os.listdir(out_path)) == 0:
        shutil.rmtree(out_path)
      else:
        return Dataset(out_path, read_only=True)
    # ====== download the file ====== #
    if os.path.exists(zip_path) and override:
      os.remove(zip_path)
    if not os.path.exists(zip_path):
      get_file(name, path, DataLoader.BASE_DIR)
    # ====== upzip dataset ====== #
    unzip_folder(zip_path, out_path, remove_zip=True)
    ds = Dataset(out_path, read_only=True)
    md5_checksum = clazz.md5(ext=ext)
    if md5_checksum is not None:
      assert ds.md5 == md5_checksum, "MD5 checksum mismatch for dataset: %s" % ds.path
    return ds

# ===========================================================================
# Images dataset
# ===========================================================================
class MNIST(DataLoader):
  pass

class MNIST_original(DataLoader):
  pass

class FMNIST_original(DataLoader):
  pass

class MNIST_dropout(DataLoader):
  pass

class FMNIST_dropout(DataLoader):
  pass

class CIFAR10(DataLoader):
  pass

class CIFAR100(DataLoader):
  pass

# ===========================================================================
# AUdio dataset
# ===========================================================================
class TIDIGITS(DataLoader):
  """ Audio digits dataset
  Encrypted and required password for access
  """

  @classmethod
  def md5(clazz, ext=''):
    return '82e2d8df7c376341a1f6deb35acdd1b0c569f0af82fb3f0173' + \
    'd11af2b74780a6b27611fa1ad0aaf16f0d4f52bddc9be1640ac3881f1' + \
    'ad074ce34f59650702632f919301507166bc190620ef168d5ef89b983' + \
    'a428e38814f17af05a4bc51b9a913d1300d54239a6cfa7dd5d75bfdc1' + \
    '7d2d958e5c3b50b3da499cf5c5eab9ee574'

  @classmethod
  def get_dataset(clazz, ext='', override=False):
    # ====== all path ====== #
    name = clazz.get_name(ext) + '.zip'
    path = base64.decodebytes(DataLoader.ORIGIN).decode() + name
    zip_path = clazz.get_zip_path(ext)
    out_path = clazz.get_ds_path(ext)
    # ====== check out_path ====== #
    if os.path.isfile(out_path):
      raise RuntimeError("Found a file at path: %s, we need a folder "
                         "to unzip downloaded files." % out_path)
    elif os.path.isdir(out_path):
      if override or len(os.listdir(out_path)) == 0:
        shutil.rmtree(out_path)
      else:
        return Dataset(out_path, read_only=True)
    # ====== download the file ====== #
    if os.path.exists(zip_path) and override:
      os.remove(zip_path)
    if not os.path.exists(zip_path):
      get_file(name, path, DataLoader.BASE_DIR)
    # ====== upzip dataset ====== #
    unzip_aes(in_path=zip_path, out_path=out_path)
    ds = Dataset(out_path, read_only=True)
    if ds.md5 != clazz.md5():
      ds.close()
      shutil.rmtree(out_path)
      raise RuntimeError("Incorrect password for loading DIGITS dataset")
    else:
      os.remove(zip_path)
    return ds

class SPEECH_SAMPLES(DataLoader):

  @classmethod
  def get_dataset(clazz, ext='', override=False):
    # ====== all path ====== #
    name = clazz.get_name(ext) + '.zip'
    path = base64.decodebytes(DataLoader.ORIGIN).decode() + name
    zip_path = clazz.get_zip_path(ext)
    out_path = clazz.get_ds_path(ext)
    # ====== check out_path ====== #
    if os.path.isfile(out_path):
      raise RuntimeError("Found a file at path: %s, we need a folder "
                         "to unzip downloaded files." % out_path)
    elif os.path.isdir(out_path):
      if override or len(os.listdir(out_path)) == 0:
        shutil.rmtree(out_path)
      else:
        return Dataset(out_path, read_only=True)
    # ====== download the file ====== #
    if os.path.exists(zip_path) and override:
      os.remove(zip_path)
    if not os.path.exists(zip_path):
      get_file(name, path, DataLoader.BASE_DIR)
    # ====== upzip dataset ====== #
    unzip_aes(in_path=zip_path, out_path=out_path)
    ds = Dataset(out_path, read_only=True)
    if os.path.exists(zip_path):
      os.remove(zip_path)
    return ds

class FSDD(object):
  """ Free Spoken Digit Dataset
  A simple audio/speech dataset consisting of recordings of
  spoken digits in wav files at 8kHz. The recordings are
  trimmed so that they have near minimal silence at the
  beginnings and ends.

  Return
  ------
  list : contain path to all .wav files of the dataset
  numpy.ndarray : the numpy array of loaded `metadata.csv`

  Reference
  ---------
  Link: https://github.com/Jakobovski/free-spoken-digit-dataset
  """
  LINK = 'https://github.com/Jakobovski/free-spoken-digit-dataset/archive/v1.0.7.zip'

  @classmethod
  def load(clazz):
    """ Return
    records: list of all path to recorded audio files
    metadata: numpy.ndarray
    """
    dat_path = get_datasetpath(name='FSDD', override=False)
    tmp_path = dat_path + '_tmp'
    zip_path = dat_path + '.zip'
    # ====== download zip dataset ====== #
    if not os.path.exists(dat_path) or \
    len(os.listdir(dat_path)) != 1501:
      if not os.path.exists(zip_path):
        get_file(fname='FSDD.zip', origin=FSDD.LINK, outdir=get_datasetpath())
      if os.path.exists(tmp_path):
        shutil.rmtree(tmp_path)
      unzip_folder(zip_path=zip_path, out_path=tmp_path, remove_zip=True)
      tmp_path = os.path.join(tmp_path, os.listdir(tmp_path)[0])
      # ====== get all records ====== #
      record_path = os.path.join(tmp_path, 'recordings')
      all_records = [os.path.join(record_path, i)
                     for i in os.listdir(record_path)]
      for f in all_records:
        name = os.path.basename(f)
        shutil.copy2(src=f, dst=os.path.join(dat_path, name))
      # ====== copy the metadata ====== #
      meta_path = os.path.join(tmp_path, 'metadata.py')
      import imp
      meta = imp.load_source('metadata', meta_path).metadata
      assert len(set(len(i) for i in meta.values())) == 1, "Invalid metadata"
      rows = []
      for name, info in meta.items():
        info = sorted(info.items(), key=lambda x: x[0])
        header = ['name'] + [i[0] for i in info]
        rows.append([name] + [i[1] for i in info])
      with open(os.path.join(dat_path, 'metadata.csv'), 'w') as f:
        for r in [header] + rows:
          f.write(','.join(r) + '\n')
    # ====== clean ====== #
    if os.path.exists(tmp_path):
      shutil.rmtree(tmp_path)
    # ====== return dataset ====== #
    all_files = [os.path.join(dat_path, i)
                 for i in os.listdir(dat_path)
                 if '.wav' in i]
    meta = np.genfromtxt(os.path.join(dat_path, 'metadata.csv'),
                         dtype=str, delimiter=',')
    return all_files, meta

  @classmethod
  def get_dataset(clazz):
    """ Return
    records: list of all path to recorded audio files
    metadata: numpy.ndarray
    """
    return clazz.load()

# ===========================================================================
# More experimental dataset
# ===========================================================================
class IRIS(DataLoader):
  pass

# ===========================================================================
# Speech synthesis
# ===========================================================================
class CMUarctic(DataLoader):
  pass

# ===========================================================================
# Others
# ===========================================================================
class MUSAN(DataLoader):
  pass

class openSMILEsad(DataLoader):
  """ This dataset contains 2 files:
  * lstmvad_rplp18d_12.net
  * rplp18d_norm.dat
  """
  pass

# ===========================================================================
# Others
# ===========================================================================
def load_glove(ndim=100):
  """ Automaticall load a MmapDict which contains the mapping
      (word -> [vector])
  where vector is the embedding vector with given `ndim`.
  """
  ndim = int(ndim)
  if ndim not in (50, 100, 200, 300):
    raise ValueError('Only support 50, 100, 200, 300 dimensions.')
  link = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL2dsb3ZlLjZCLiVkZA==\n'
  link = str(base64.decodebytes(link) % ndim, 'utf-8')
  fname = os.path.basename(link)
  embedding = get_file(fname, link, outdir=get_datasetpath(root='~'))
  return MmapDict(embedding, read_only=True)

def load_lre_sad():
  """
  key: 'LDC2017E23/data/eval/lre17_lqoyrygc.sph'
  value: [(1.99, 3.38), (8.78, 16.41)] (in second)
  """
  link = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL2xyZV9zYWQ=\n'
  link = str(base64.decodebytes(link), 'utf-8')
  path = get_file(fname=os.path.basename(link),
                  origin=link,
                  outdir=get_datasetpath(root='~'))
  return MmapDict(path=path, read_only=True)

def load_lre_list():
  """ The header include following column:
  * name: LDC2017E22/data/ara-acm/ar-20031215-034005_0-a.sph
  * lre: {'train17', 'eval15', 'train15', 'dev17', 'eval17'}
  * language: {'ara-arb', 'ara-ary', 'ara-apc', 'ara-arz', 'ara-acm',
               'eng-gbr', 'eng-usg', 'eng-sas',
               'fre-hat', 'fre-waf'
               'zho-wuu', 'zho-cdo', 'zho-cmn', 'zho-yue', 'zho-nan',
               'spa-lac', 'spa-eur', 'spa-car',
               'qsl-pol', 'qsl-rus',
               'por-brz'}
  * corpus: {'pcm', 'alaw', 'babel', 'ulaw', 'vast', 'mls14'}
  * duration: {'3', '30', '5', '15', '10', '20', '1000', '25'}

  Note
  ----
  Suggested namming scheme:
    `lre/lang/corpus/dur/base_name`
  """
  link = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL2xyZV9saXN0LnR4dA==\n'
  link = str(base64.decodebytes(link), 'utf-8')
  path = get_file(fname=os.path.basename(link),
                  origin=link,
                  outdir=get_datasetpath(root='~'))
  return np.genfromtxt(fname=path, dtype=str, delimiter=' ',
                       skip_header=1)

def load_sre_list():
  link = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL1NSRV9GSUxFUy56aXA=\n'
  link = str(base64.decodebytes(link), 'utf-8')
  ds_path = get_datasetpath(name='SRE_FILES', root='~',
                            is_folder=False, override=False)
  if os.path.exists(ds_path) and len(os.listdir(ds_path)) != 24:
    shutil.rmtree(ds_path)
  if not os.path.exists(ds_path):
    path = get_file(fname=os.path.basename(link),
                    origin=link,
                    outdir=get_datasetpath(root='~'))
    unzip_folder(zip_path=path, out_path=ds_path, remove_zip=True)
  return Dataset(ds_path, read_only=True)

def load_voxceleb_list():
  link = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLWRhdGFzZXRzL3ZveGNlbGViX2xpc3RzLnppcA==\n'
  link = str(base64.decodebytes(link), 'utf-8')
  ds_path = get_datasetpath(name='voxceleb_lists', root='~',
                            is_folder=False, override=False)
  if not os.path.exists(ds_path):
    path = get_file(fname=os.path.basename(link),
                    origin=link,
                    outdir=get_datasetpath(root='~'))
    unzip_folder(zip_path=path, out_path=os.path.dirname(path),
                 remove_zip=True)
  return Dataset(ds_path, read_only=True)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = 'float32,gpu'

import numpy as np

from odin import backend as K, nnet as N, visual as V
from odin import preprocessing as pp
from odin.utils import (args_parse, stdio,
                        get_module_from_path, get_script_path)
from odin.utils.mpi import cpu_count

from utils import (WAV_FILES, SAMPLED_WAV_FILE,
                   PATH_ACOUSTIC_FEAT, PATH_EXP)
# ===========================================================================
# Config
# ===========================================================================
stdio(os.path.join(PATH_EXP, 'features_extraction.log'))
args = args_parse(descriptions=[
    ('recipe', 'the name of function defined in feature_recipes.py', None),
    ('--debug', 'enable debug or not', None, False)
])
DEBUG = args.debug
# ===========================================================================
# Create the recipes
# ===========================================================================
extractor = get_module_from_path(identifier=str(args.recipe),
                                 prefix='feature_recipes',
                                 path=get_script_path())
assert len(extractor) > 0, \
"Cannot find any recipe with name: '%s' from path: '%s'" % (args.recipe, get_script_path())
recipe = extractor[0](DEBUG)
# ====== debugging ====== #
if DEBUG:
  with np.warnings.catch_warnings():
    np.warnings.filterwarnings('ignore')
    for path, name in SAMPLED_WAV_FILE:
      feat = recipe.transform(path)
      assert feat['bnf'].shape[0] == feat['mspec'].shape[0]
      V.plot_multiple_features(feat, title=feat['name'])
    V.plot_save(os.path.join(PATH_EXP, 'features_%s.pdf' % args.recipe))
    exit()
# ===========================================================================
# Prepare the processor
# ===========================================================================
with np.warnings.catch_warnings():
  np.warnings.filterwarnings('ignore')
  jobs = list(WAV_FILES.keys())
  processor = pp.FeatureProcessor(jobs=jobs,
      path=os.path.join(PATH_ACOUSTIC_FEAT, args.recipe),
      extractor=recipe,
      n_cache=1200,
      ncpu=min(18, cpu_count() - 2),
      override=True,
      identifier='name',
      log_path=os.path.join(PATH_EXP, 'processor_%s.log' % args.recipe),
      stop_on_failure=False)
  processor.run()
  pp.validate_features(processor,
                       nb_samples=12,
                       path=os.path.join(PATH_EXP, args.recipe),
                       override=True)



from __future__ import print_function, division, absolute_import

import numpy as np

import tensorflow as tf

from tensorflow.python.keras import Model, Sequential
from tensorflow.python.keras.layers import Dense, Layer

from tensorflow_probability.python.distributions import (
  softplus_inverse, Dirichlet)

from odin.bay.distribution_layers import DirichletLayer, OneHotCategoricalLayer
from odin.bay.helpers import kl_divergence

__all__ = [
    'LatentDirichletAllocation'
]


class LatentDirichletAllocation(Model):
  """ Variational Latent Dirichlet Allocation

  To maintain good intuition behind the algorithm, we name the
  attributes as for topics discovery task in natural language
  processing.

  Parameters
  ----------
  n_components : int, optional (default=10)
    Number of topics in LDA.

  components_prior : float (default=0.7)
    the topic prior concentration for Dirichlet distribution

  References
  ----------
  [1]: David M. Blei, Andrew Y. Ng, Michael I. Jordan. Latent Dirichlet
       Allocation. In _Journal of Machine Learning Research_, 2003.
       http://www.jmlr.org/papers/volume3/blei03a/blei03a.pdf
  [2]: Michael Figurnov, Shakir Mohamed, Andriy Mnih. Implicit Reparameterization
       Gradients, 2018
       https://arxiv.org/abs/1805.08498
  [3]: Akash Srivastava, Charles Sutton. Autoencoding Variational Inference For
       Topic Models. In _International Conference on Learning Representations_,
       2017.
       https://arxiv.org/abs/1703.01488
  """

  def __init__(self, n_components=10, components_prior=0.7,
               encoder_layers=[64, 64], activation='relu',
               n_mcmc_samples=1, analytic_kl=True,
               random_state=None):
    super(LatentDirichletAllocation, self).__init__()
    self._random_state = np.random.RandomState(seed=random_state) \
        if not isinstance(random_state, np.random.RandomState) \
          else random_state
    self._initializer = tf.initializers.GlorotNormal(
        seed=self._random_state.randint(1e8))

    self.n_components = int(n_components)
    self.components_prior = np.array(softplus_inverse(components_prior))

    self.n_mcmc_samples = n_mcmc_samples
    self.analytic_kl = analytic_kl
    # ====== encoder ====== #
    encoder = Sequential(name="Encoder")
    for num_hidden_units in encoder_layers:
      encoder.add(
          Dense(num_hidden_units,
                activation=activation,
                kernel_initializer=self._initializer))
    encoder.add(
        Dense(n_components,
              activation=tf.nn.softplus,
              kernel_initializer=self._initializer,
              name="DenseConcentration"))
    encoder.add(DirichletLayer(clip_for_stable=True,
                               pre_softplus=False,
                               name="topics_posterior"))
    self.encoder = encoder
    # ====== decoder ====== #
    # The observations are bag of words and therefore not one-hot. However,
    # log_prob of OneHotCategorical computes the probability correctly in
    # this case.
    self.decoder = OneHotCategoricalLayer(
      probs_input=True, name="bag_of_words")

  def build(self, input_shape):
    n_features = input_shape[1]
    # decoder
    self.topics_words_logits = self.add_weight(
        name="topics_words_logits",
        shape=[self.n_components, n_features],
        initializer=self._initializer)
    # prior
    self.prior_logit = self.add_weight(
        name="prior_logit",
        shape=[1, self.n_components],
        trainable=False,
        initializer=tf.initializers.Constant(self.components_prior))
    # call this to set built flag to True
    super(LatentDirichletAllocation, self).build(input_shape)

  def call(self, inputs):
    docs_topics_posterior = self.encoder(inputs)
    docs_topics_samples = docs_topics_posterior.sample(self.n_mcmc_samples)

    # [n_topics, n_words]
    topics_words_probs = tf.nn.softmax(self.topics_words_logits, axis=1)
    # [n_docs, n_words]
    docs_words_probs = tf.matmul(docs_topics_samples, topics_words_probs)
    output_dist = self.decoder(
      tf.clip_by_value(docs_words_probs, 1e-4, 1 - 1e-4))

    # initiate prior, concentration is clipped to stable range
    # for Dirichlet
    concentration = tf.clip_by_value(
        tf.nn.softplus(self.prior_logit), 1e-3, 1e3)
    topics_prior = Dirichlet(
        concentration=concentration, name="topics_prior")

    # ELBO
    kl = kl_divergence(q=docs_topics_posterior, p=topics_prior,
                       use_analytic_kl=self.analytic_kl,
                       q_sample=self.n_mcmc_samples,
                       auto_remove_independent=True)
    if self.analytic_kl:
      kl = tf.expand_dims(kl, axis=0)
    llk = output_dist.log_prob(inputs)
    ELBO = llk - kl

    # maximizing ELBO, hence, minizing following loss
    self.add_loss(tf.reduce_mean(-ELBO))
    self.add_metric(tf.reduce_mean(kl), aggregation='mean', name="MeanKL")
    self.add_metric(tf.reduce_mean(-llk), aggregation='mean', name="MeanNLLK")

    return output_dist
